#include "Matcher.h"
#include "io.h"
#include <algorithm>
#include <limits>
#include <sstream>
#include <unordered_map>
#include <iostream>
#include <stdexcept>

struct Token_Labels {
  struct pair {
    std::size_t start_idx = std::numeric_limits<size_t>::max();
    std::size_t end_idx = std::numeric_limits<size_t>::min();
  };
  std::unordered_map<std::string, int> token_map;
  std::vector<Camera> camera_labels_sorted;
  std::vector<pair> pairs;
  int biggest_camera_id = -1;
};

Token_Labels token_labels;
Token_Labels token_labels_fuzzy;
Token_Labels token_labels2;

bool initialized = false;
int total_labels_amount;

static Token_Labels create_int_labels(const std::vector<Camera> &labels) {
  int token_id = 0;
  int camera_id = 0;
  Token_Labels token_labels_local{};

  std::set<std::string> unique;
  std::string unique_id;

  for (auto cam : labels) {
    camera_id = std::max(camera_id, cam.model_id);
    for (auto &word : cam.toke_set) {
      if (!token_labels_local.token_map.count(word)) {
        token_labels_local.token_map.insert({word, token_id});
        token_id++;
      }
      auto pos = token_labels_local.token_map.find(word);
      cam.label.push_back(pos->second);
    }
    std::sort(begin(cam.label), end(cam.label));
    unique_id.clear();
    for (int number : cam.label) {
      unique_id += std::to_string(number);
    }
    // if label is unique
    if (unique.find(unique_id) == unique.end()) {
      token_labels_local.camera_labels_sorted.push_back(cam);
      unique.insert(unique_id);
    }
  }
  sort(
          begin(token_labels_local.camera_labels_sorted),
          end(token_labels_local.camera_labels_sorted),
          [](const Camera &a, const Camera &b) { return a.label[0] < b.label[0]; });
  token_labels_local.pairs = std::vector<Token_Labels::pair>(token_id);
  for (size_t i = 0; i < token_labels_local.camera_labels_sorted.size(); ++i) {
    if (token_labels_local.pairs[token_labels_local.camera_labels_sorted[i].label[0]]
                .start_idx == std::numeric_limits<size_t>::max()) {
      token_labels_local.pairs[token_labels_local.camera_labels_sorted[i].label[0]].start_idx = i;
      token_labels_local.pairs[token_labels_local.camera_labels_sorted[i].label[0]].end_idx = i + 1;
    } else {
      token_labels_local.pairs[token_labels_local.camera_labels_sorted[i].label[0]].end_idx++;
    }
  }
  token_labels_local.biggest_camera_id = camera_id;
  return token_labels_local;
}

static Token_Labels create_fuzzy_labels(const std::vector<Camera> &labels, std::set<std::string>& can_remove, size_t size) {
  std::vector<Camera> labels_fuzzy;
  labels_fuzzy.reserve(labels.size());
  if(size > 1000){
    can_remove.insert("alpha");
    can_remove.insert("1");
  }

  for (auto &cam : labels) {
    Camera c = cam;
    c.toke_set.clear();
    std::set_difference(begin(cam.toke_set), end(cam.toke_set),
                        begin(can_remove), end(can_remove),
                        std::inserter(c.toke_set, begin(c.toke_set)));
    if (!c.toke_set.empty() && c.toke_set.size() < cam.toke_set.size()) {
      labels_fuzzy.emplace_back(c);
    }
  }
  return create_int_labels(labels_fuzzy);
}

static std::vector<int> title_to_int_label(const std::string &title, std::unordered_map<std::string, int> &token_map) {
  std::vector<int> int_label{};
  std::stringstream ss(title);
  std::string word;
  std::set<int> toke_set;
  while (ss >> word) {
    auto pos = token_map.find(word);
    if (pos != token_map.end()) {
      toke_set.insert(pos->second);
    }
  }
  int_label.reserve(toke_set.size());
  for (auto int_token : toke_set) {
    int_label.push_back(int_token);
  }
  sort(begin(int_label), end(int_label));
  return int_label;
}

static bool contains_all(const std::vector<int> &title_tokens, const std::vector<int> &label,
                         size_t i = 0) {
  if (title_tokens.size() - i < label.size())
    return false;
  size_t j = 0;
  size_t count = 0;
  while (i < title_tokens.size() && j < label.size()) {
    if (title_tokens[i] < label[j])
      i++;
    else if (label[j] < title_tokens[i])
      j++;
    else {
      count++;
      i++;
      j++;
    }
  }
  return count == label.size();
}

static std::vector<Camera> try_to_match(const std::vector<int> &title_int_label, const Token_Labels &token_labels_local) {
  std::vector<Camera> vlabels;

  for (size_t i = 0; i < title_int_label.size(); ++i) {
    int token = title_int_label[i];
    const size_t start_idx = token_labels_local.pairs[token].start_idx;
    const size_t end_idx = token_labels_local.pairs[token].end_idx;
    for (size_t j = start_idx; j < end_idx; ++j) {
      if (contains_all(title_int_label, token_labels_local.camera_labels_sorted[j].label, i)) {
        vlabels.push_back(token_labels_local.camera_labels_sorted[j]);
      }
    }
  }

  sort(begin(vlabels), end(vlabels), [](const Camera &a, const Camera &b) {
    return a.label.size() > b.label.size();
  });
  // unique labels, no subsets, unique model_id's
  if (vlabels.size() < 16) { // if to much possible labels than ignore
    for (int64_t i = 0; i < int64_t(vlabels.size() - 1); ++i) {
      for (size_t j = i + 1; j < vlabels.size(); ++j) {
        if (contains_all(vlabels[i].label, vlabels[j].label) || vlabels[i].model_id == vlabels[j].model_id ) {
          vlabels.erase(begin(vlabels) + j);
          --j;
        }
      }
    }
  }
  return vlabels;
}

void Matcher::init(const std::string& path_labels) {
  if(!initialized){
    // can remove must be loaded first
    std::set<std::string> can_remove;
    for(const auto& cam : read_labels(path_labels + "can_remove.txt")){
      for(const auto& s : cam.toke_set){
        can_remove.insert(s);
      }
    }

    auto labels = read_labels(path_labels + "labels_a.txt");
    token_labels = create_int_labels(labels);

    auto labels2 = read_labels(path_labels + "labels_b.txt", token_labels.biggest_camera_id + 1);
    token_labels_fuzzy = create_fuzzy_labels(labels, can_remove, labels2.size());
    token_labels2 = create_int_labels(labels2);

    total_labels_amount = 1 + token_labels2.biggest_camera_id;
    initialized = true;

  } else {
    throw std::runtime_error("matcher already initialized");
  }
}

int Matcher::size() {
  if(!initialized){
    throw std::runtime_error("uninitialized matcher");
  } else {
    return total_labels_amount;
  }
}

std::vector<Camera> Matcher::match(const Record& record) {
  if(initialized) {
    auto title_int_label = title_to_int_label(record.title, token_labels.token_map);
    auto cameras = try_to_match(title_int_label, token_labels);
    // save here all cameras found during matching
    std::vector<Camera> cameras_found;
    if (cameras.empty()) {
      auto title_int_label_fuzzy = title_to_int_label(record.title, token_labels_fuzzy.token_map);
      auto cameras_fuzzy = try_to_match(title_int_label_fuzzy, token_labels_fuzzy);
      if (!cameras_fuzzy.empty()) {
        for(const auto& cam : cameras_fuzzy){
          cameras_found.push_back(cam);
        }
      } else {
        auto title_int_label2 = title_to_int_label(record.title, token_labels2.token_map);
        auto cameras2 = try_to_match(title_int_label2, token_labels2);
        for (const auto &cam : cameras2) {
          cameras_found.push_back(cam);
        }
      }
    } else {
      for(const auto& cam : cameras){
        cameras_found.push_back(cam);
      }
    }
    return cameras_found;
  } else {
    throw std::runtime_error("uninitialized matcher");
  }
}






